@@ -31,7 +31,9 @@
 #define BANNER "\nThis is dvidvi 1.0, Copyright (C) 1988-91, Radical Eye Software\n"
 #define STRINGSIZE (500)  /* maximum number of strings in program */
 
+#include "dvidvi.h"
 #ifndef VMS
+#include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
 #else /* VMS */
@@ -136,6 +138,12 @@ integer *pageloc ;
 integer *pagenumbers ;
 int prettycolumn ;       /* the column we are at when running pretty */
 
+// To handle xxx1 for headers
+// Lets suppose there will be less than 256 of those.
+char *headers[256];
+integer ln_headers[256];
+integer nb_headers = 0L;
+
 /*
  *   This array holds values that indicate the length of a command, if
  *   we aren't concerned with that command (which is most of them) or
@@ -289,15 +297,19 @@ integer p;
 void stringdvibuf(p,n)
 integer p,n;
 {
+  char *s = temp;
   fseek(infile,p,SEEK_SET);
   while ( n-- > 0 )
-    *temp++ = dvibyte();
+    *s++ = dvibyte();
 }
 
 /*
  *   Print a usage error messsage, and quit.
  */
-usage() {
+#ifdef __GNUC__
+__attribute__((noreturn))
+#endif
+void usage() {
    (void)fprintf(stderr,banner);
    (void)fprintf(stderr,"Usage:  dvidvi [options] input[.dvi] [output]\n");
    (void)fprintf(stderr,"where options are:\n");
@@ -317,8 +329,7 @@ usage() {
 /*
  *   Print an error message, and exit if it is fatal.
  */
-error(s)
-char *s ;
+void error(char *s)
 {
    (void)fprintf(stderr, "%s\n", s) ;    /* AKT: was dvidvi: %s */
    if (*s == '!')
@@ -367,7 +378,7 @@ integer whole, num, den, sf ;
  *
  *   (Could blow up if a parameter * mag / 1000 > 2^30 sp.)
  */
-scalemag(p)
+void scalemag(p)
 long *p ;
 {
    int negative ;
@@ -538,11 +549,12 @@ int i;
  *   Parse the arguments to the routine, and stuff everything away
  *   into those globals above.
  */
+void
 processargs(argc, argv)
 int argc ;
 char *argv[] ;
 {
-   char *p, *q ;
+   char *q ;
    int i, pageno, lastext = -1 ;
    long hoffset, voffset ;
    int reversed ;
@@ -620,9 +632,9 @@ case 'l':
 	       p++; lasttransf = 1 ;
 	       }
 #ifdef SHORTINT
-            if(sscanf(p, "%ld", &lastpage)==0)
-#else   /* ~SHORTINT */
             if(sscanf(p, "%d", &lastpage)==0)
+#else   /* ~SHORTINT */
+            if(sscanf(p, "%ld", &lastpage)==0)
 #endif  /* ~SHORTINT */
                error("! Bad last page option (-l).") ;
             break ;
@@ -684,9 +696,9 @@ case 'n' :
             if (*p == 0 && argv[i+1])
                p = argv[++i] ;
 #ifdef SHORTINT
-            if (sscanf(p, "%ld", &maxpages)==0)
-#else   /* ~SHORTINT */
             if (sscanf(p, "%d", &maxpages)==0)
+#else   /* ~SHORTINT */
+            if (sscanf(p, "%ld", &maxpages)==0)
 #endif  /* ~SHORTINT */
                error("! Bad number of pages option (-n).") ;
             break ;
@@ -835,11 +847,12 @@ void searchpageloc()
    (void)dvibuf(p);
    pagenumbers[num]=signedquad();
 }
+
 /*
  *   This routine simply reads the entire dvi file, and then initializes
  *   some values about it.
  */
-readdvifile() {
+void readdvifile() {
    integer p ;
    unsigned char c,d,e ;
 
@@ -909,6 +922,7 @@ readdvifile() {
  */
 
    p = pageloc[0L] + 45 ;
+nextcmd:
    c=dvibuf(p);
    while (comlen[c]) {
       p += comlen[c] ;
@@ -920,11 +934,35 @@ readdvifile() {
         landscape = p ;
         rem0special = 1 ;
       }
+      stringdvibuf(p+2L,6L);
+      if (strncmp(temp, "header", 6)==0) {
+	char l1;
+	l1 = dvibuf(p+1);
+	stringdvibuf(p+2,l1);
+	if ( nb_headers == 256 ) {
+	   fprintf(stderr,"\n"
+			  "There are more than 256 PS headers in your DVI file.\n"
+			  "Please ask a wizzard to enlarge me :-)\n");
+	}
+	ln_headers[nb_headers] = l1;
+	headers[nb_headers++] = strdup(temp);
+	/*
+	 * Seek at the end of the xxx1, then goto reading next commands.
+	 * In a normal DVI file, as produced by LaTeX, all of those special
+	 * commands (xxx1 with PS headers) are outputed on the very first
+	 * page, all together. So a simple loop might be enough. Just in
+	 * case we have a strange file we accept that there might be some
+	 * typesetting commandes between those xxx1.
+	 */
+	p = p + l1 + 2;
+	goto nextcmd;
+      }
    }
 }
 /*
  *   Output a single byte, keeping track of where we are.
  */
+void
 outdvibyte(c)
 unsigned char c ;
 {
@@ -934,6 +972,7 @@ unsigned char c ;
 /*
  *   Send out two bytes.
  */
+void
 outdvi2(v)
 integer v ;
 {
@@ -943,6 +982,7 @@ integer v ;
 /*
  *   Send out a longword.
  */
+void
 outdviquad(v)
 integer v ;
 {
@@ -953,6 +993,7 @@ integer v ;
  *   This routine just copies some stuff from the buffer on out.
  *   Suppose the file is positioned correctly before
  */
+void
 putbuf(length)
 integer length ;
 {
@@ -962,6 +1003,7 @@ integer length ;
 /*
  *   This routine outputs a string, terminated by null.
  */
+void
 putstr(s)
 register unsigned char *s ;
 {
@@ -971,6 +1013,7 @@ register unsigned char *s ;
 /*
  *   Here we write the preamble to the dvi file.
  */
+void
 writepreamble() {
 /*   just copy the first 14 bytes of the file */
    fseek(infile,0L,SEEK_SET);
@@ -981,6 +1024,7 @@ writepreamble() {
 /*
  *   This routine writes out a font definition.
  */
+void
 putfontdef(f)
 int f ;
 {
@@ -994,6 +1038,7 @@ int f ;
 /*
  *   The postamble is next.
  */
+void
 writepostamble() {
    int i ;
    integer p ;
@@ -1019,6 +1064,7 @@ writepostamble() {
 /*
  *   This routine starts a page, by writing out a bop command.
  */
+void
 beginpage() {
    int i ;
    integer p ;
@@ -1035,10 +1081,11 @@ beginpage() {
  *   This routine sends out a page.  We need to handle the
  *   landscape special, though.
  */
+void
 dopage(num)
 integer num ;
 {
-   register integer p,q ;
+   register integer p ;
    register int len ;
    integer v, oldp ;
    unsigned char c;
@@ -1063,15 +1110,40 @@ integer num ;
       }
       prettycolumn += i + 1 ;
 #ifdef SHORTINT
-      (void)fprintf(stderr, "[%ld", num+1) ;
-#else  /* ~SHORTINT */
       (void)fprintf(stderr, "[%d", num+1) ;
+#else  /* ~SHORTINT */
+      (void)fprintf(stderr, "[%ld", num+1) ;
 #endif /* ~SHORTINT */
       (void)fflush(stderr) ;
    }
    p = pageloc[num] + 45 ;
    c=dvibuf(p);
    while (c != 140) {
+      if ( nb_headers ) {
+	 integer i;
+	 /*
+	  * Need to output the 'nb_headers' headers that we collected.
+	  * If we are outputing the real first page of the previous
+	  * DVI file, then the headers will be doubled, which is not
+	  * troublesome: dvips will keep only one of those.
+	  */
+	 for ( i=0; i<nb_headers; i++ ) {
+	    outdvibyte(239);
+	    outdvibyte(ln_headers[i]);
+	    putstr(headers[i]);
+	    if ( !quiet ) {
+	       fprintf(stderr,"<%s>",headers[i]+7);
+	       prettycolumn += ln_headers[i]+2-7;
+	    }
+	 }
+	 /*
+	  * We will output the headers only one time. That is possible that
+	  * those headers are not usefull (the pages where they are used might
+	  * not be kept in the out file, but there is no simple way to avoid
+	  * it.
+	  */
+         nb_headers = 0;
+      }
       if ((len=comlen[c]) > 0) {    /* most commands are simple */
          outdvibyte(c);
          putbuf((long)len-1) ;
@@ -1136,6 +1208,7 @@ default:       fprintf(stderr, "Bad dvi command was %d at %ld\n", len, p) ;
 /*
  *   Here we end a page.  Simple enough.
  */
+void
 endpage() {
    outputpages++ ;
    outdvibyte(140) ;
@@ -1144,12 +1217,12 @@ endpage() {
  *   This is our main routine for output, which runs through all the
  *   pages we need to output.
  */
+void
 writedvifile() {
    integer pagenum ;
    int ppp ;
    integer actualpageno ;
    struct pagespec *ps ;
-   integer p ;
    Boolean beginp ;
 
    writepreamble() ;
@@ -1191,6 +1264,7 @@ writedvifile() {
    }
    writepostamble() ;
 }
+int
 main(argc, argv)
 int argc ;
 char *argv[] ;
@@ -1198,5 +1272,9 @@ char *argv[] ;
    processargs(argc, argv) ;
    readdvifile() ;
    writedvifile() ;
+   if (!quiet) {
+      fprintf(stderr,"\n");
+   }
+   return 0;
 }
 
